/* Domination checking routines */

# include <stdio.h>
# include <stdlib.h>
# include <math.h>

# include "global.h"
# include "rand.h"

int check_pareto_dominance_guided2d(individual *a, individual *b);
int check_pareto_dominance_guided3d(individual *a, individual *b);
int check_par_dom(individual *a, individual *b);

/* Routine for non-domination checking (Pareto and equitable efficiency).
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */

int check_dominance (individual *a, individual *b)
{
    if (a->constr_violation<0 && b->constr_violation<0)
    {
        if (a->constr_violation > b->constr_violation)
        {
            return (1);
        }
        else
        {
            if (a->constr_violation < b->constr_violation)
            {
                return (-1);
            }
            else
            {
                return (0);
            }
        }
    }
    else
    {
        if (a->constr_violation < 0 && b->constr_violation == 0)
        {
            return (-1);
        }
        else
        {
            if (a->constr_violation == 0 && b->constr_violation <0)
            {
                return (1);
            }
            else
            {
            	switch (efficiency)
				{
            		case 0: return check_pareto_dominance(a, b);
            		case 1: return check_equitable_dominance(a, b);
            		case 2: return check_variable_dominance_minimum(a, b);
            		case 3: return check_variable_dominance_nondomination(a, b);
            		default: exit (1);
				}
            }
        }
    }
}




/* Routine that actually checks Pareto dominance.
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_pareto_dominance(individual *a, individual *b)
{
	if (restr==4)
	{
		if (nobj==2)
		{
			return check_pareto_dominance_guided2d(a, b);
		}
		else if (nobj==3)
		{
			return check_pareto_dominance_guided3d(a, b);
		}
		else
		{
			printf("Number of objectives must be 2 or 3!");
			exit(1);
		}
	}
	else
	{
		return check_par_dom(a, b);
	}
}

/* It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_par_dom(individual *a, individual *b)
{
	int i;
	int flag1 = 0;
	int flag2 = 0;
	for (i=0; i<nobj; ++i)
	{
		if (a->obj[i] < b->obj[i])
		{
			flag1 = 1;

		}
		else
		{
			if (a->obj[i] > b->obj[i])
			{
				flag2 = 1;
			}
		}
	}
	if (flag1==1 && flag2==0)
	{
		return (1);
	}
	else
	{
		if (flag1==0 && flag2==1)
		{
			return (-1);
		}
		else
		{
			return (0);
		}
	}
}

/* It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_pareto_dominance_guided2d(individual *a, individual *b)
{
	static double mat2d[2][2] = {{1, 0}, {0.75, 1}}; /*{{11, 12}, {21, 22}};*/
	int flag = 0;
	double omega1_b, omega2_b;
	double omega1_a, omega2_a;
	double bak1_a, bak2_a;
	double bak1_b, bak2_b;

	omega1_a = mat2d[0][0]*a->obj[0] + mat2d[0][1]*a->obj[1];
	omega2_a = mat2d[1][0]*a->obj[0] + mat2d[1][1]*a->obj[1];
	bak1_a = a->obj[0]; /* store original objective function values */
	bak2_a = a->obj[1];
	a->obj[0] = omega1_a; /* set modified objective function values */
	a->obj[1] = omega2_a;

	omega1_b = mat2d[0][0]*b->obj[0] + mat2d[0][1]*b->obj[1];
	omega2_b = mat2d[1][0]*b->obj[0] + mat2d[1][1]*b->obj[1];
	bak1_b = b->obj[0]; /* store original objective function values */
	bak2_b = b->obj[1];
	b->obj[0] = omega1_b; /* set modified objective function values */
	b->obj[1] = omega2_b;

	flag = check_par_dom(a, b);

	/* restore original objective function values */
	a->obj[0] = bak1_a;
	a->obj[1] = bak2_a;
	b->obj[0] = bak1_b;
	b->obj[1] = bak2_b;

	return flag;
}

/* It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_pareto_dominance_guided3d(individual *a, individual *b)
{
	static double mat2d[3][3] = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}; /*{{11, 12, 13}, {21, 22, 23}, {31, 32, 33}}*/
	int flag = 0;
	double omega1_b, omega2_b, omega3_b;
	double omega1_a, omega2_a, omega3_a;
	double bak1_a, bak2_a, bak3_a;
	double bak1_b, bak2_b, bak3_b;

	omega1_a = mat2d[0][0]*a->obj[0] + mat2d[0][1]*a->obj[1] + mat2d[0][2]*a->obj[2];
	omega2_a = mat2d[1][0]*a->obj[0] + mat2d[1][1]*a->obj[1] + mat2d[0][2]*a->obj[2];
	omega3_a = mat2d[2][0]*a->obj[0] + mat2d[2][1]*a->obj[1] + mat2d[2][2]*a->obj[2];
	bak1_a = a->obj[0]; /* store original objective function values */
	bak2_a = a->obj[1];
	bak3_a = a->obj[2];
	a->obj[0] = omega1_a; /* set modified objective function values */
	a->obj[1] = omega2_a;
	a->obj[2] = omega3_a;

	omega1_b = mat2d[0][0]*b->obj[0] + mat2d[0][1]*b->obj[1] + mat2d[0][2]*b->obj[2];
	omega2_b = mat2d[1][0]*b->obj[0] + mat2d[1][1]*b->obj[1] + mat2d[0][2]*b->obj[2];
	omega3_b = mat2d[2][0]*b->obj[0] + mat2d[2][1]*b->obj[1] + mat2d[2][2]*b->obj[2];
	bak1_b = b->obj[0]; /* store original objective function values */
	bak2_b = b->obj[1];
	bak3_b = b->obj[2];
	b->obj[0] = omega1_b; /* set modified objective function values */
	b->obj[1] = omega2_b;
	b->obj[2] = omega3_b;

	flag = check_par_dom(a, b);

	/* restore original objective function values */
	a->obj[0] = bak1_a;
	a->obj[1] = bak2_a;
	a->obj[2] = bak3_a;
	b->obj[0] = bak1_b;
	b->obj[1] = bak2_b;
	b->obj[2] = bak3_b;

	return flag;
}


/* Routine that actually checks equitable dominance.
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_equitable_dominance(individual *a, individual *b)
{
	int i, j;
	int flag1 = 0;
	int flag2 = 0;
	double temp;
	double *obj_arr_a = (double *) malloc(nobj*sizeof(double));
	double *obj_arr_b = (double *) malloc(nobj*sizeof(double));

	for(i=0; i<nobj; ++i)
	{
		obj_arr_a[i] = a->obj[i];
		obj_arr_b[i] = b->obj[i];
	}

	/* build ascending ordering maps */
	quicksort_arr(obj_arr_a, 0, nobj-1);
	quicksort_arr(obj_arr_b, 0, nobj-1);

	/* reverse ordering maps */
	for (i=0, j=nobj-1; i<j; ++i, --j)
	{
		temp = obj_arr_a[j];
		obj_arr_a[j] = obj_arr_a[i];
		obj_arr_a[i] = temp;
		temp = obj_arr_b[j];
		obj_arr_b[j] = obj_arr_b[i];
		obj_arr_b[i] = temp;
	}

	/* build cumulative ordering maps */
	for (i=1; i<nobj; ++i)
	{
		obj_arr_a[i] += obj_arr_a[i-1];
		obj_arr_b[i] += obj_arr_b[i-1];
	}

	/* check domination */
	for (i=0; i<nobj; ++i)
	{
		if (obj_arr_a[i] < obj_arr_b[i])
		{
			flag1 = 1;

		}
		else
		{
			if (obj_arr_a[i] > obj_arr_b[i])
			{
				flag2 = 1;
			}
		}
	}


	free(obj_arr_a);
	free(obj_arr_b);

	if (flag1==1 && flag2==0)
	{
		return (1);
	}
	else
	{
		if (flag1==0 && flag2==1)
		{
			return (-1);
		}
		else
		{
			return (0);
		}
	}

}

/* Routine that actually checks variable domination (minimum).
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_variable_dominance_minimum(individual *a, individual *b)
{
	int j;
	double d1, d2;
	double *d = malloc(nobj*sizeof(double));
	double *e = malloc(nobj*sizeof(double));

	d1 = 0;
	for (j=0; j<nobj; ++j)
	{
		d[j] = a->obj[j] - b->obj[j];
		e[j] = a->obj[j] - ideal[j];
		d1 += d[j] * e[j];
	}
	d2 = gamma * array_norm(d, nobj) * array_minimum(e, nobj);
	if (d1 > d2)
	{
		free(d);
		free(e);
		return (-1);
	}

	d1 = 0;
	for (j=0; j<nobj; ++j)
	{
		d[j] = b->obj[j] - a->obj[j];
		e[j] = b->obj[j] - ideal[j];
		d1 += d[j] * e[j];
	}
	d2 = gamma * array_norm(d, nobj) * array_minimum(e, nobj);
	free(d);
	free(e);
	if (d1 > d2)
	{
		return (1);
	}
	else
	{
		return (0);
	}
}

/* Routine that actually checks variable domination (nondomination).
   It will return the following values
   1 if a dominates b
   -1 if b dominates a
   0 if both a and b are non-dominated */
int check_variable_dominance_nondomination(individual *a, individual *b)
{
	int j;
	double d1, d2;
	double *d = malloc(nobj*sizeof(double));
	double *e = malloc(nobj*sizeof(double));

	d1 = 0;
	for (j=0; j<nobj; ++j)
	{
		d[j] = a->obj[j] - b->obj[j];
		e[j] = b->obj[j] - ideal[j];
		d1 += d[j] * e[j];
	}
	d2 = gamma * array_norm(d, nobj) * array_minimum(e, nobj);
	if (d1 > d2)
	{
		free(d);
		free(e);
		return (-1);
	}

	d1 = 0;
	for (j=0; j<nobj; ++j)
	{
		d[j] = b->obj[j] - a->obj[j];
		e[j] = a->obj[j] - ideal[j];
		d1 += d[j] * e[j];
	}
	d2 = gamma * array_norm(d, nobj) * array_minimum(e, nobj);
	free(d);
	free(e);
	if (d1 > d2)
	{
		return (1);
	}
	else
	{
		return (0);
	}
}



